// *************************************************************************
//  Collection.h
// *************************************************************************

#ifndef _COLLECTION_H_
#define _COLLECTION_H_

#include <stdlib.h>
#include <assert.h>

#include "Rule.h"

// *************************************************************************
//
//  This class replaces the collection struct which was painfully
//  obfuscated.
//
//  This structure is basically a dynamic Array of pointers to Clauses (rules).
//  As an extra, it can have a Symbol Table inside it to group all rules
//  with the same name togather in lists.  This ability is currently
//  used only within magic.
//
//  To do this grouping of rules, the 'link' field of the Clause is used,
//  otherwise it should be empty (NULL).  Obviously then, a Clause should
//  be in at most one collection.
//  Unlike the previous version of Collection, you can turn the hash
//  table grouping on and off while using the Collection. 
// 
//  It is important that the rules be kept in the same order they were
//  inserted, as this structure is iterated through to print out the
//  rules in order (like for the rewriting phase).
//
// *************************************************************************

    // The initial size of the Collection array (it can grow)
#define    COR_DEFLT_COLLCTN_SIZE    32

// *************************************************************************

class Cor_CArg;
class Cor_Clause;
class Cor_SymTable;
class Cor_CArgStack;

// *************************************************************************

class Cor_Collection {

  private:

    Cor_Clause    **rules;    // The dynamic array of Clauses
    int             length;   // The length of the array (that's being used)
    int             size;     // The current size of the rules array;

    int             hash_on;  // Will be set if hashing is currently turned on
    Cor_SymTable   *table;    // The Symbol Table, if hash_on is set

    Cor_CArgStack  *param_slots;  // This is non-null if there are parameters
                                  // to fill in (only for C++ Interface)

  public:

    Cor_Collection(int init_size=0, int init_hash_on=0);
   ~Cor_Collection();

    int  count() { return(length); }
    int  is_empty() { return(length == 0); }

      // To get a _copy_ of the ith element
    Cor_Clause *get_ith(int i) {
        if (i >= length)
            return(NULL);
        return(rules[i]);
    }

      // If hashing is on, return the first Clause that has this name,
      // the others are obtained by following the Clause's link field.
      // If hashing is off, this always returns NULL (not found).
    Cor_Clause *get_match(Cor_CArg& name);

      // To turn hashing on and off (this is idempotent)
    int use_hash_table(int on=1);

      // push() returns the position of the inserted rule
    int  push(Cor_Clause *rule);

        // Pull out the last element
    Cor_Clause *pop() {
        if (length > 0) {
            length--;
            Cor_Clause *tmp = rules[length];

              // This is just so no one else try's grabbing this one
            rules[length] = NULL;
            return(tmp);
        } else {
            return(NULL);
        }
    }

      // deletes all stored rules and re-initializes the structure
    void empty();

      // A few hacks for magic which wants to selectively delete rules
      // Warning: this leaves holes in the array, use squash() to get
      //     rid of them.
    void del_ith(int i) {
       
        assert(hash_on == 0); 

        if (i < length) {
            delete rules[i];
            rules[i] = NULL;
        }
    }

      // After a few del_ith's, your array will have holes, this is
      // bad, so this function compacts it.
      // It returns the number of holes removed
    int squash();
};

// *************************************************************************

//  A dynamic array to hold multiple connections

class Cor_CollectStack {

  public:
    Cor_Collection **data;
    int              length;
    int              data_size;

    Cor_CollectStack() {
        data = NULL;
        length = 0;
        data_size = 0;
    }

   ~Cor_CollectStack() {
        empty();
        delete [] data;
    }

    int  count() { return(length); }
    int  is_empty() { return(length == 0); }

      // push() returns the position of the inserted collection 
    int  push(Cor_Collection *col);

    Cor_Collection *get_ith(int i) {
        if (length > i)
            return(data[i]);
        return(NULL);
    }

        // Pull out the last element
    Cor_Collection *pop() {
        if (length > 0) {
            length--;
            Cor_Collection *tmp = data[length];

              // This is just so no one else try's grabbing this one
            data[length] = NULL;
            return(tmp);
        } else {
            return(NULL);
        }
    }

      // Empty out the array - delete each element
    void empty();
};

// *************************************************************************

#endif _COLLECTION_H_
