// *************************************************************************
//    PersistentRelation.C
// *************************************************************************

#ifdef WITH_PERSISTENCE

#include <string.h>

#include "PersistentRelation.h"
#include "Interface.h"
#include "Relation.h"
#include "RelationStruct.h"
#include "RelationList.h"
#include "TupleIterator.h"
#include "Profile.h"
#include "RuleInfo.h"
#include "Annotations.h"
#include "Table.h"
#include "DatabaseStruct.h"
#include "Debug.h"

// *************************************************************************

extern int SM_initialized;
extern Cor_DatabaseStruct *Cor_ExodusDB;

#define DELETE_INDEX_ENTRY 1
#define DELETE_OBJ_AND_INDEX_ENTRY 2

// *************************************************************************

extern int Cor_unify_literal(Cor_TupleIterator&, Cor_Tuple *, int);

// *************************************************************************
//  Initialize the PersistentRelation structure from the catalog if it exists
//  in the catalog.  Otherwise, an error is raised.
//
//  Note that the file names are only guarenteed to be unique up to
//  COR_MAX_FILE_NAME_LEN

Cor_PersistentRel::Cor_PersistentRel(Cor_CArg& name, int arity, int& error)
    : Cor_StorageRelation(arity, -1), index_set(this)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized)
        return;

    OID oid;

    if (!SM.lookupFile(name, arity, &oid)) {
        *(Cor_exEnv->err_strm) << 
            "SORRY ! Unknown persistent relation referenced !\n";
        error = 1;
        return;
    }

    USERDESC *u_desc;
    if (!SM.retrieveObject(&oid, &u_desc)) {
        *(Cor_exEnv->err_strm) << 
            "SORRY ! Unknown persistent relation referenced !\n";
        *(Cor_exEnv->err_strm) << "Error in catalogs !!\n";
        error = 1;
        return;
    }

    fd.copyDesc((Cor_FileDescStruct *)(u_desc->basePtr));

    this->name = name;
    this->oid  = oid;
    ref_count = 1;              // This is to prevent it from being emptied
                                // when it is closed in a database other than
                                // db_rels.

    lddesc= NULL;
    count = fd.count;

    check_subsum = !fd.duplicates;
    r_kind = COR_R_EXODUS;
    SM.releaseObject(u_desc);

    int index_offset = -1;
    char attrbuf[COR_MAX_ARGLIST_SIZE+1];

    // create appropriate in-memory index structures
    for (int i = 0; i < fd.num_indexes; i++) {
        if ((index_offset = fd.nextIndex(index_offset)) < 0) {
            CORAL_error(COR_INDEX_ERR, NULL, "PersistentRel::Ct");
            return;
        }
        for (int j = 0; j < arity; j++) {
            attrbuf[j] = ((fd.indexes[index_offset].attrs[j]) ? '1' : '0');
        }
        attrbuf[arity] = '\0';

        if (this->add_index(attrbuf) == COR_I_FAIL) {
            *(Cor_exEnv->err_strm)
                << "Failed to add index to persistent relation "
                << this->name.symbol() << endl;
         }
    }

    // enter into the exodus workspace
    if (!Cor_ExodusDB->RelationTable)
        Cor_ExodusDB->RelationTable = new Cor_SymTable();

    
    Cor_ExodusDB->RelationTable->insert(name, (Pointer) this);
}

// *************************************************************************
Cor_PersistentRel::Cor_PersistentRel(Cor_CArg& name, int arity, int& error,
        OID *oidptr) : Cor_StorageRelation(arity, -1), index_set(this)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

    oid = *oidptr;

    USERDESC *u_desc;
    if (!SM.retrieveObject(&oid, &u_desc)) {
        *(Cor_exEnv->err_strm) << 
            "SORRY ! Unknown persistent relation referenced !\n";
        *(Cor_exEnv->err_strm) << "Error in catalogs !!\n";
        error = 1;
        return;
    }

    fd.copyDesc((Cor_FileDescStruct *)(u_desc->basePtr));

    this->name = name;
    this->oid  = oid;
    ref_count = 1;              // This is to prevent it from being emptied
                                // when it is closed in a database other than
                                // db_rels.

    lddesc = NULL;
    r_kind = COR_R_EXODUS;
    count = fd.count;
    SM.releaseObject(u_desc);

    int index_offset = -1;
    char attrbuf[COR_MAX_ARGLIST_SIZE+1];

    // create appropriate in-memory index structures
    for (int i = 0; i < fd.num_indexes; i++) {
        if ((index_offset = fd.nextIndex(index_offset)) < 0) {
            CORAL_error(COR_INDEX_ERR, NULL, "PersistentRel::Ct");
            return;
        }
        for (int j = 0; j < arity; j++) {
            attrbuf[j] = ((fd.indexes[index_offset].attrs[j]) ? '1' : '0');
        }
        attrbuf[arity] = '\0';

        if (this->add_index(attrbuf) == COR_I_FAIL) {
            *(Cor_exEnv->err_strm)
                << "Failed to add index to persistent relation "
                << this->name.symbol() << endl;
         }
    }

    // enter into the exodus workspace
    if (!Cor_ExodusDB->RelationTable)
        Cor_ExodusDB->RelationTable = new Cor_SymTable();

    Cor_ExodusDB->RelationTable->insert(name, (Pointer) this);
}

// *************************************************************************
//
//  This version of the constructor is solely used for temporary relations
//  used during module computation, and does not insert the relation into
//  the persistent workspace table.

Cor_PersistentRel::Cor_PersistentRel(Cor_CArg& name, int arity, int& error,
            Cor_FileDescStruct *fdptr)
    : Cor_StorageRelation(arity, -1), index_set(this)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized)
        return;

    fd.copyDesc(fdptr);

    this->name = name;
    INVALIDATE_OID((this->oid));
    ref_count = 1;              // This is to prevent it from being emptied
                                // when it is closed in a database other than
                                // db_rels.
    lddesc = NULL;
    count = fd.count;
    r_kind = COR_R_EXODUS;

    int index_offset = -1;
    char attrbuf[COR_MAX_ARGLIST_SIZE+1];

    // create appropriate in-memory index structures
    for (int i = 0; i < fd.num_indexes; i++) {
        if ((index_offset = fd.nextIndex(index_offset)) < 0) {
            CORAL_error(COR_INDEX_ERR, NULL, "PersistentRel::Ct");
            return;
        }
        for (int j = 0; j < arity; j++) {
            attrbuf[j] = ((fd.indexes[index_offset].attrs[j]) ? '1' : '0');
        }
        attrbuf[arity] = '\0';

        if (this->add_index(attrbuf) == COR_I_FAIL) {
            *(Cor_exEnv->err_strm)
                << "Failed to add index to persistent relation "
                << this->name.symbol() << endl;
        }
    }
}

// *************************************************************************
void propagateChangesToDisk()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized)
        return;

    Cor_SymCursor *cursor = NULL;

    if (Cor_ExodusDB && Cor_ExodusDB->RelationTable)
        cursor = Cor_ExodusDB->RelationTable->OpenScan();

    if (cursor) {

        do {
            Cor_PersistentRel *p_rel = (Cor_PersistentRel *) cursor->data();
            if (p_rel)
                p_rel->flush_to_disk();
 
        } while(cursor->ScanNext());
    }

    delete cursor;
}

// *************************************************************************
// Send the info in the PersistentRel back on disk

void Cor_PersistentRel::flush_to_disk()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    fd.count = count;

    USERDESC *u_desc;
    if (!SM.retrieveObject(&oid, &u_desc)) {
        *(Cor_exEnv->err_strm) << 
            "SORRY ! Unknown persistent relation referenced !\n";
        *(Cor_exEnv->err_strm) << "Error in catalogs !!\n";
        return;
    }

    // write out the new file descriptors
    int e = sm_WriteObject(SM.bufGroup(), 0, sizeof(Cor_FileDescStruct),
        (char *)(&fd), u_desc, TRUE );
    ErrorCheck(e, "sm_WriteObject");
    
}

// *************************************************************************
void Cor_PersistentRel::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name.symbol() << "/" <<  arity()
         << "  : (exodus)  ";
    *out << name.symbol() << "/" << arity() << "  : (exodus)  ";
    index_set.print_adornments(out);
//  fd.print_index_adornments(out);
    *out << ": count " << count << endl;
    *out << endl;
}

// *************************************************************************
void Cor_PersistentRel::print_facts(ostream *out, Cor_ArgList *) const
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << "rel_open(" << name.symbol() << "(";
    for (int i = 0; i < arity() -1; i++) {
        *out << "x,";
    }

    if (arity())
        *out << "x), db_rels).\n";
    else
        *out << "), db_rels).\n";
}

// *************************************************************************
//  Inserts the tuple object specified by the arglist into the relation.
//
//  The schema describes the structure of the data.
//  The entire object is first constructed in memory, and a persistent
//  object is created out of it at one shot. In the case
//  of strings, the length of the string(an int) is first written to the
//  file, followed by the actual string. The object is also added to the
//  appropriate existing indexes.

void Cor_PersistentRel::insertObject(Cor_ArgList &arglist)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized)
        return;

    int e = 0;

    // Check that arity of relation and arglist match - defensive programming
    if (arglist.arity() != arity()) {
        CORAL_error(COR_BAD_ARGLIST, "this error shouldn't happen",
            "PersistentRel::insert");
        return;
    }

    int obj_len = 0;

    char **data_ptr = new char* [arity()];    // array of pointers to data items
    int  *data_size = new int [arity()];

    // determine size of each field of the object
    for (int i = 0; i < arity(); i++) {
        switch(fd.schema[i]) {
        case COR_STRING_DBTYPE:
            data_size[i] = strlen(data_ptr[i] = make_string(arglist[i])) + 1;
            // to store the length of the string along with the string
            obj_len += data_size[i] * sizeof(char) + sizeof(int); 
            break;
        case COR_INT_DBTYPE:
            obj_len += sizeof(int);
            break;
        case COR_DOUBLE_DBTYPE:
            obj_len += sizeof(double);
            break;
        default:
            // This should not happen !!
            CORAL_error(COR_BAD_SCHEMA,
                "INVALID field reached", "insertObject");
            return;
        }
    }

    char *obj_data = new char[obj_len];
    int obj_offset = 0;
    int tmp_int;
    double tmp_double;

    // create an in-memory copy of the object data
    for (int i = 0; i < arity(); i++) {
        switch(fd.schema[i]) {
        case COR_STRING_DBTYPE:
            memcpy(obj_data+obj_offset,(char *)(data_size + i), sizeof(int));
            obj_offset += sizeof(int);
            memcpy(obj_data+obj_offset,(char *)(data_ptr[i]), 
             data_size[i]*sizeof(char));
            obj_offset += data_size[i]*sizeof(char);
            delete data_ptr[i];
            break;
        case COR_INT_DBTYPE:
            tmp_int = make_int(arglist[i]);
            data_ptr[i] = (char *)(&tmp_int);
            memcpy((char *)(obj_data)+obj_offset,(char *)(data_ptr[i]),
                sizeof(int));
            obj_offset += sizeof(int);
            break;
        case COR_DOUBLE_DBTYPE:
            tmp_double = make_double(arglist[i]);
            data_ptr[i] = (char *)(&tmp_double);
            memcpy(obj_data+obj_offset,(char *)(data_ptr[i]), sizeof(double));
            obj_offset += sizeof(double);
            break;
        default:
            // This should not happen !!
            CORAL_error(COR_BAD_SCHEMA,
                "INVALID field reached", "insertObject");
            return;
        }

        ASSERT((obj_offset <= obj_len));
    }

    delete [] data_ptr;
    delete [] data_size;


    // now create the new object using the buffer group of the Storage Manager!
    // This is probably not optimal, but its better than creating a separate 
    // buffer pool just for this
    OID oid;

    if (lddesc) {
        e = sm_LoadNextObject(lddesc, obj_len, (void *)obj_data, &oid);
        if (ErrorCheck(e, "sm_LoadNextObject")) {
            delete [] obj_data;
            return;
        }
    } else {
        e = sm_CreateObject(SM.bufGroup(), &(fd.fid), NEAR_LAST, NULL, NULL,
            obj_len, (void *)obj_data, &oid);
        if (ErrorCheck(e, "sm_CreateObject")) {
            delete [] obj_data;
            return;
        }
    }

    delete [] obj_data;

    USERDESC *u_desc;
    int index_offset = -1;

    // Add object to all valid indexes
    if (fd.num_indexes) {

        if (!SM.retrieveObject(&oid, &u_desc)) {
            *(Cor_exEnv->err_strm) << 
                "SORRY ! Unknown persistent relation referenced !\n";
            *(Cor_exEnv->err_strm) << "Error in catalogs !!\n";
            return;
        }

        for (int i = 0; i < fd.num_indexes; i++) {

            if ((index_offset = fd.nextIndex(index_offset)) < 0) {
                CORAL_error(COR_INDEX_ERR, NULL, "PersistentRel::insertObject");
                SM.releaseObject(u_desc);
                return;
            }
            fd.addObjectToIndex(index_offset, u_desc);
        }
    }

    (fd.count)++;

    if (fd.num_indexes)
        SM.releaseObject(u_desc);
}

// *************************************************************************
//  Determine if the retrieved tuple satisfies the bindings so far.
//  If so, add to the bindings(change the bindenv) and return
//  To do this, create an arglist that reflects the tuple
//  and call unify_arglists(iterator.arglist, new_arglist,iterator.bindenv,
//      NULL_BINDENV);
//  This automatically checks for validity of the tuple and if valid
//  also pushes the new environment on to the stack.
//
//  In general, the tuple may contain variable length strings as fields.
//  So the length of a string field is stored as an integer in front of it.

Cor_Tuple *Cor_PersistentRel::checkTuple(USERDESC *u_desc,
            Cor_TupleIterator &iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return NULL;

    Cor_ArgList *new_arglist;
    Cor_StackMark  stackmark0;
    int offset = 0;
    int str_len = 0;
    int data_len = 0;

    // Create an arglist corresponding to the retrieved tuple.
    new_arglist = new Cor_ArgList(this->arity());

    for (int i=0; i < arity(); i++) {
        switch(fd.schema[i]) {
        case COR_STRING_DBTYPE:
            memcpy((char *)&str_len ,(char *)(u_desc->basePtr+offset),
                sizeof(int));
            offset = offset + sizeof(int);
            data_len = str_len * sizeof(char);
            if ((!((iterator.arg_list)[i]).is_var())
                || (! ((iterator.arg_list)[i]).nm_equals(*Cor_DontCareSymbol)))
            {
                Cor_CArg *carg = make_arg((char *)(u_desc->basePtr+offset));
                (*new_arglist)[i] = *carg;
                carg->destroy();
            }
            else {
                (*new_arglist)[i] = iterator.arg_list[i];
            }
            break;
    
        case COR_INT_DBTYPE:
            data_len = sizeof(int);
            int itmp;
            if ((!((iterator.arg_list)[i]).is_var())
               || (! ((iterator.arg_list)[i]).nm_equals(*Cor_DontCareSymbol)))
            {
                memcpy((char *)&itmp, (char *)(u_desc->basePtr+offset),
                    sizeof(int));
                (*new_arglist)[i] = Cor_CArg(itmp);
            }
            else {
                (*new_arglist)[i] = iterator.arg_list[i];
            }
            break;
    
        case COR_DOUBLE_DBTYPE :
            data_len = sizeof(double);
            double dtmp;
            if ((!((iterator.arg_list)[i]).is_var())
                || (! ((iterator.arg_list)[i]).nm_equals(*Cor_DontCareSymbol)))
            {
                memcpy((char *)&dtmp, (char *)(u_desc->basePtr+offset),
                    sizeof(double));
                (*new_arglist)[i] = Cor_CArg(dtmp);
            }
            else {
                (*new_arglist)[i] = iterator.arg_list[i];
            }
            break;
    
        default:
            CORAL_error(COR_BAD_SCHEMA, "INVALID field reached", "checkTuple");
            return 0;
        }
        offset += data_len;
    }

    // Create a Tuple corresponding to the retrieved tuple.
    Cor_Tuple *tuple = new Cor_Tuple(new_arglist);

    // Call unify_literal
    if (Cor_unify_literal(iterator, tuple, Cor_exEnv->dbg_indexing))
        return tuple;

    for (int ti = 0; ti < tuple->arity(); ti++) {
        if (tuple->arg(ti).is_num()) {
            // LEAK delete tuple->arg(ti);
        }
    }
    delete new_arglist;
    delete tuple;

    stackmark0.pop_to();
    return NULL;
}

// *************************************************************************
//  Gets the next tuple from the persistent relation using a file scan.
//  The tuple should satisfy the selection implicit in the arglist of
//  the iterator.  The new bindings due to the selected tuple are
//  (automatically) pshed onto the global trail stack.

#define FAIL_RETURN \
  { \
    iterator.scanDesc = NULL; \
    iterator.ipos = Scanning_Done; \
    iterator.set_no_match(); \
    return NULL; \
  }

Cor_Tuple *Cor_PersistentRel::fileScanGetNext(Cor_TupleIterator &iterator,
        int delete_flag)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) FAIL_RETURN;

    int e;
    USERDESC *u_desc;
    BOOL eof;

  // if (exEnv->C_interrupt_raised) FAIL_RETURN;
  if (Cor_exEnv->C_interrupt_raised) {
    iterator.ipos = Scanning_Done;
    iterator.set_no_match();
    return NULL; 
  }
    // If this is the first time get_next is being called, the scan
    // may have to be initialized. However, subsequent get_next() calls
    // should continue from where they left off.

    if (!iterator.scanDesc) {
        e = sm_OpenScanWithGroup(&(fd.fid), PHYSICAL_SCAN, SM.bufGroup(),
            (SCANDESC **)&(iterator.scanDesc), NULL);
        if (ErrorCheck(e, "sm_OpenScanWithGroup"))
            FAIL_RETURN;
    }

    for (;;) {
        e = sm_ScanNextObject((SCANDESC *)iterator.scanDesc, 0, READ_ALL,
            &u_desc, &eof);
        if (ErrorCheck(e, "sm_ScanNextObject") || eof) {
            // If there are no more tuples, return.
            e = sm_CloseScan((SCANDESC *)iterator.scanDesc);
            ErrorCheck(e, "sm_CloseScan");
            FAIL_RETURN;
        }

        // Determine if the retrieved tuple satisfies the bindings so far.
        // If so, add to the bindings(change the bindenv) and return
        Cor_Tuple *tuple;
        if ((tuple = checkTuple(u_desc, iterator))) {
            if (delete_flag) {
                SM.destroyObject(&(u_desc->oid), SM.bufGroup());
            }
            return tuple;
        }
        // Else, get next tuple (return to loop )
    }
}

#undef FAIL_RETURN

// *************************************************************************
//  Gets the next tuple from the persistent relation using an index scan
//  using the given index id. The tuple should satisfy the selection
//  implicit in the arglist of the iterator. The new bindings due to
//  the selected tuple are (automatically) pushed onto the global
//  trail stack.

#define FAIL_RETURN \
  { \
    INVALIDATE_CURSOR(*((SMCURSOR *)iterator.cursor)); \
    iterator.ipos = Scanning_Done; \
    iterator.set_no_match(); \
    return NULL; \
  }

Cor_Tuple *Cor_PersistentRel::indexScanGetNext(Cor_TupleIterator &iterator,
        int index, int delete_flag)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) FAIL_RETURN;

    USERDESC *u_desc;
    BOOL eof;
    KEY key;
    OID oid;
    int tmp_int;
    double tmp_double;
    char value_buff[COR_MAX_KEY_LEN];
    int buff_offset = 0;
    char *data_ptr = NULL;
    int data_len;
    int str_len;
    int e;

    if (Cor_exEnv->C_interrupt_raised) FAIL_RETURN;

    // If this is the first time get_next is being called, the scan
    // may have to be initialized. However, subsequent get_next() calls
    // should continue from where they left off.

    if (CURSOR_IS_INVALID(*((SMCURSOR *)iterator.cursor))) {

        // Fill value_buff with the key value.
        // This is done by copying the appropriate attributes into value_buff

        ASSERT(iterator.arg_list.count() == arity());

        for (int i = 0; i < arity(); i++) {

            Cor_Term term(iterator.arg_list[i], iterator.bindenv);
            Cor_full_dereference_term(term);

            if (!(term.expr.isConstant()))
                continue;

            switch(fd.schema[i]) {
            case COR_INT_DBTYPE:
                if (!term.expr.is_num()) {
                    *(Cor_exEnv->err_strm) << "Error in scan of relation "
                                           <<  name.symbol() << "\n";
                    CORAL_error(COR_BAD_ARGLIST, "int expected",
                        "indexScanGetNext");
                    term.expr.printon(Cor_exEnv->err_strm);      
                    FAIL_RETURN
                }
                tmp_int = make_int(term.expr);
                data_ptr = (char *)&tmp_int;
                data_len = sizeof(int);
                break;

            case COR_STRING_DBTYPE:
                if (!term.expr.is_symbol()) {
                    *(Cor_exEnv->err_strm)
                        << "Error in scan of relation "
                        << name.symbol() << endl;
                    CORAL_error(COR_BAD_ARGLIST, "string expected",
                        "indexScanGetNext");
                    term.expr.printon(Cor_exEnv->err_strm);
                    FAIL_RETURN
                }
                data_ptr = make_string(term.expr);
                str_len = strlen(data_ptr)+1;
                data_len = str_len * sizeof(char);
                break;

            case COR_DOUBLE_DBTYPE:
                if (!term.expr.is_num()) {
                    *(Cor_exEnv->err_strm)
                        << "Error in scan of relation "
                        << name.symbol() << endl;
                    CORAL_error(COR_BAD_ARGLIST, "double expected",
                        "indexScanGetNext");
                    term.expr.printon(Cor_exEnv->err_strm);      
                    FAIL_RETURN
                }
                tmp_double = make_double(term.expr);
                data_ptr = (char *)&tmp_double;
                data_len = sizeof(double);
                break;

            default:
                // This should not happen !!
                CORAL_error(COR_BAD_SCHEMA, "INVALID field reached",
                    "indexScanGetNext");
                FAIL_RETURN;
            }

            if ((buff_offset+data_len < COR_MAX_KEY_LEN) &&
                fd.indexes[index].attrs[i])
            {
                memcpy((char *)(value_buff + buff_offset),(char *)(data_ptr),
                    data_len );
                buff_offset += data_len;
            }
        }

        key.length = (TWO)(buff_offset);
        key.valuePtr = (void *)(value_buff);

        // Initialize the index scan
        e = sm_FetchInit(&(fd.indexes[index].iid), SM.bufGroup(), &key, SM_EQ,
            &key, SM_EQ, (SMCURSOR *)iterator.cursor);
        
        if (ErrorCheck(e, "sm_FetchInit"))
            FAIL_RETURN;
    }

    // At this stage, the scan has already been initialized either during
    // this function invocation, or in some previous invocation.
    for (;;) {
        // Get the next OID from the index
        e = sm_FetchNext((SMCURSOR *)iterator.cursor, NULL, &oid, &eof);
        // If there are no more index entries, return.
        if (ErrorCheck(e, "sm_FetchNext") || eof) FAIL_RETURN;

        verboseError = 0;
        //if (!SM.retrieveObject(&oid, &u_desc)) FAIL_RETURN;
        if (!SM.retrieveObject(&oid, &u_desc)) continue;
        verboseError = !Cor_exEnv->C_quiet_mode_default;
        
        Cor_Tuple *tuple;
        // Determine if the retrieved tuple satisfies the bindings so far.
        // If so, if delete_flag is set, delete the object, else
        // add to the bindings(change the bindenv) and return
        if ((tuple = checkTuple(u_desc, iterator))) {
            if (delete_flag)  {
                // remove the index entry for the object
                (fd.indexes[index]).delete_entry(u_desc, fd.schema);
                if (delete_flag == DELETE_OBJ_AND_INDEX_ENTRY)
                    SM.destroyObject(&(u_desc->oid), SM.bufGroup());
            }
            SM.releaseObject(u_desc);
            return tuple;
        }

        // Else, get next tuple (return to loop )
        SM.releaseObject(u_desc);
    }
}

#undef FAIL_RETURN

// *************************************************************************
void Cor_PersistentRel::initialize_iterator(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return;

#    ifdef DEBUG
        if (Cor_exEnv->dbg_get_next) {
            *(Cor_exEnv->trace_strm) << 
                "\nPersistentRelation::initialize_iterator: ";
            iterator.relation->name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ( ";
            iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " )\n";
        }
#    endif

    iterator.stack_mark.get_mark();

    if (!delta_indexed) {
        if (iterator.start_mark) {
            //    don't need to go to disk
            iterator.ipos = Scanning_New;
            iterator.ppos = iterator.start_mark->tindex;
        } else {
              // need to scan from disk
              // find the appropriate index to be used
            iterator.ipos = Scanning_Known;
            iterator.index = index_set.find_index(iterator);
        }
    } else {
        if (!iterator.in_delta()) {
              // Starting on main relation
            if (iterator.start_mark) {
                  // go to the delta relations
                iterator.ipos = Scanning_Init;
                iterator.delta_rel_pos =  (iterator.start_mark)->rel_p;
            } else {
                  // need to scan from disk
                  // find the appropriate index to be used
                iterator.ipos = Scanning_Known;
                iterator.index = index_set.find_index( iterator );
            }
        } else {
              // In a delta relation
            iterator.index = index_set.find_index( iterator );
        }
    }

    if (iterator.index) {
#        ifdef DEBUG
            if (Cor_exEnv->dbg_indexing) {
                *(Cor_exEnv->trace_strm) << "Index scan chosen"
                    << "[ btree index "
                    << ((Cor_PersistentIndex *)(iterator.index))->chosen_index
                        << " ]\n";
            }
#        endif
        // Index scan cursor needs to be created
        iterator.cursor = (void *) new SMCURSOR;
        INVALIDATE_CURSOR(*((SMCURSOR *)iterator.cursor));
    }
}

// *************************************************************************
Cor_BindEnv *Cor_PersistentRel::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return NULL;

    this->get_next_tuple(iterator);
    if (iterator.no_match()) return NULL;
    return(iterator.bindenv);
}


// *************************************************************************
#define FAIL_RETURN \
  { \
    iterator.ipos = Scanning_Done;     \
    iterator.set_no_match();    \
    return NULL;            \
  }

Cor_Tuple *Cor_PersistentRel::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_PROFILE
        if (Cor_T_Stack->count > 0) {
          // NEW_CHNG:wX
          (Cor_T_Stack->cur_md()->
                module_info->Cor_GlobalProfileInfo.getnexts)++;

          //((T_Stack.top()->moduledata->
          //        module_info->GlobalProfileInfo.getnexts)++;
        }
        else {
            (Cor_GlobalProfileInfo.getnexts)++;
        }
#    endif
    
    iterator.tuple_env = NULL;
    Cor_Tuple *result = NULL;

    if (Cor_exEnv->C_interrupt_raised) FAIL_RETURN;

    if (!SM_initialized) {
        *(Cor_exEnv->err_strm) << "Exodus not initialized!\n";
        FAIL_RETURN;
    }

    // initialize
    if (iterator.ipos == Scanning_Init)
        initialize_iterator(iterator);
    else
        iterator.stack_mark.pop_to();

    if (!delta_indexed) {
        if (iterator.ipos != Scanning_New) {
            //    need to scan from disk
            // If no index is available, use a full file scan.
            if (!iterator.index) {
                // This sets iterator.scanDesc
                result = fileScanGetNext(iterator);
            }
            else {
                // iterator.index is set
                result = (iterator.index)->get_next_tuple(iterator);
            }

            if (result != NULL) {
                // found something in main disk relation
                delete result->_args;
                delete result;
                return(NULL); 
            }
            else {
                iterator.ipos = Scanning_New;
                iterator.bkt  = &tlist;
                iterator.ppos = 0;
            }
        }

        // need to scan from the chain now
        Cor_StackMark stack_mark;
        Cor_Tuple *tuple;

        while (iterator.ppos < tlist.count()) {

            if (iterator.end_mark
                    && (iterator.end_mark->tindex <= iterator.ppos)) {
                tuple = NULL;
                break;
            }

            tuple = tlist.tuples[iterator.ppos];

            (iterator.ppos)++;

            if (!tuple || tuple->is_deleted())
                continue;

            if (Cor_unify_literal(iterator, tuple, Cor_exEnv->dbg_get_next))
                return(tuple);

            stack_mark.pop_to();
        }

        if (!tuple) {
            iterator.ipos = Scanning_Done;
            iterator.set_no_match();
            stack_mark.pop_to();
            return(NULL);
        }
    } else {
          // relation is delta-indexed

        if (!iterator.in_delta()) {
              // Scanning from main relation on disk
              // If no index is available, use a full file scan.
            if (!iterator.index) {
                // This sets iterator.scanDesc
                result = fileScanGetNext(iterator);
            }
            else {
                // iterator.index is set
                result = (iterator.index)->get_next_tuple(iterator);
            }

            if (result != NULL) {
                delete result->_args;
                delete result;
                return(NULL);
            }

            if (!delta_relations)
                FAIL_RETURN;

            // now go into the delta relations
            iterator.ipos = Scanning_Init;
            iterator.delta_rel_pos = delta_relations->reln_chain;
        }

        // delta_rel_pos has to have been set by this point.
        if (delta_relations != NULL) { 
            // iterator scanning deltas, this == main relation
            for(;;) { /* Loop till successful delta is found */
                if (iterator.ipos == Scanning_Init) {
                    if (iterator.delta_rel_pos == NULL)
                        FAIL_RETURN;
                    if (iterator.end_mark && (iterator.delta_rel_pos ==
                        (iterator.end_mark)->rel_p)) FAIL_RETURN;
                }
                result = iterator.delta_rel_pos->reln->get_next_tuple(iterator);
                if (result != NULL)
                    return(result);

                // Done with current delta relation.  Got to go to next delta.
                iterator.delta_rel_pos = iterator.delta_rel_pos->next;
                iterator.ipos = Scanning_Init;      
            }
        } else { 
            // iterator scanning deltas, this == delta relation
            return iterator.index->get_next_tuple(iterator);
        }
    }

    return(NULL);
}

#undef FAIL_RETURN

// *************************************************************************
void Cor_PersistentRel::freeMark(Cor_RMark *mark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (mark == NULL)
        return;

    Cor_RMark *this_mark = mark;

    // Unlink mark
    this_mark->next->prev = this_mark->prev;
    this_mark->prev->next = this_mark->next;

    // add it to the free marks list
    this_mark->next = free_marks;
    free_marks = this_mark;

    // check if this was the first mark in the list of marks
    if (this_mark->prev != &headMark) return;
        // thus, if an "intermediate" mark is deleted, nothing is done!
        // this is fine for (!delta_indexed) relations, but with delta_indexed
        // relations, an alternative is to merge the delta relations 
        // corresponding to this mark with the "previous" relation.

    if (!delta_indexed) {

        int limit;

        // Move all linked tuples upto the new first mark to hash table.
        if (there_are_marks()) {
            limit = headMark.next->tindex;
            for (Cor_RMark *hmark = headMark.next; ; hmark = hmark->next) {
                if (hmark == &headMark)
                    break;
                if (hmark->tindex != limit)
                    break;
                hmark->tlist  = &tlist;  
                hmark->tindex = 0;
            }
        } else {
            limit = tlist.count();
        }

        for (int j=0; j < limit; j++) {
            Cor_Tuple *tuple = tlist.tuples[j];

            if (!insert_raw(tuple))
                delete tuple;
            else
                count++;

            tlist.tuples[j] = NULL;
        }

        /*
        if (scan_count == 1)
            tlist.compress(0);
        */

    } else {
        Cor_RelationLink *limit;

        ASSERT((delta_relations!= NULL));
        ASSERT((delta_relations->reln_lastptr != NULL));
        
        if (there_are_marks()) 
            limit = headMark.next->rel_p;
        else
            limit = NULL;

        Cor_RelationLink *old;
        while (delta_relations->reln_chain != limit) {
            Cor_StorageRelation *reln = delta_relations->reln_chain->reln;
            old = delta_relations->reln_chain;
            delta_relations->reln_chain = old->next;

            reln->add_all_tuples(this,0,0);
            reln->empty_relation();        // Empty tuples out of relation

            old->next = free_reln_list;
            free_reln_list = old;
        }

        if (delta_relations->reln_chain == NULL)
            delta_relations->reln_lastptr = NULL;
    }
}

// *************************************************************************
void Cor_PersistentRel::free_allMarks()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RMark *hmark;
    while (headMark.next != &headMark) {
        hmark = headMark.next;
        freeMark(hmark);
    }
    ASSERT((headMark.prev == headMark.next));
}

// *************************************************************************
static void start_iterator(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return;

#    ifdef DEBUG
        if (Cor_exEnv->dbg_get_next) {
            *(Cor_exEnv->err_strm) << 
                "\nPersistentRelation::initialize_iterator: ";
            iterator.relation->name.print(NULL, Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << " ( ";
            iterator.arg_list.print(iterator.bindenv, Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << " )\n";
        }
#    endif

    iterator.stack_mark.get_mark();
    iterator.ipos = Scanning_Persistent;
}

// *************************************************************************
int Cor_PersistentRel::tuple_delete(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;
    
    Cor_Tuple *return_tuple;
    int object_deleted = 0;

    // set up index scan cursor on all available indexes
    for (int i = 0; i < COR_MAX_INDEXES; i++) {
        if (fd.indexes[i].valid()) {
            Cor_TupleIterator iterator(this, tuple);

            start_iterator(iterator);
            iterator.index = (Cor_GenericIndex *)i;
            iterator.cursor = (void *) new SMCURSOR;
            INVALIDATE_CURSOR(*((SMCURSOR *)iterator.cursor));

            if (object_deleted) {
                return_tuple = indexScanGetNext(iterator, (int)(iterator.index),
                    DELETE_INDEX_ENTRY);
            }
            else {
                return_tuple = indexScanGetNext(iterator, (int)(iterator.index),
                    DELETE_OBJ_AND_INDEX_ENTRY);
                object_deleted = 1;
            }
            iterator.stack_mark.pop_to();
            INVALIDATE_CURSOR(*((SMCURSOR *)iterator.cursor));
        }
    }

    // if the object has not been deleted yet (because no index is present),
    // it is deleted using a full file scan

    if (object_deleted) {
        (fd.count)--;
        return 1;
    }

    Cor_TupleIterator iterator(this, tuple);
    start_iterator(iterator);
    iterator.index = (Cor_GenericIndex *)(-1);

    return_tuple = fileScanGetNext(iterator, DELETE_OBJ_AND_INDEX_ENTRY);
    iterator.stack_mark.pop_to();
    int e = sm_CloseScan((SCANDESC *)iterator.scanDesc);
    ErrorCheck(e, "sm_CloseScan");
    iterator.scanDesc = NULL;
    (fd.count)--;
    return 1;
}

// *************************************************************************
int Cor_PersistentRel::is_subsumed(Cor_Tuple *tuple, Cor_RMark *, Cor_RMark *)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;
    Cor_StackMark stackmark;

    Cor_TupleIterator tuple_its(this, tuple);
    tuple_its.get_next_tuple();
    if (tuple_its.no_match()) {
        stackmark.pop_to();
        return 0;
    }
    stackmark.pop_to();
    return 1;
}

// *************************************************************************
int Cor_PersistentRel::insert_raw(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized) return 0;

    // Only flat tuples can be persistent !
#ifdef    OLD_CODE
    if (tuple->env_size) {
        CORAL_error(COR_BAD_TUPLE, "Only flat tuples can be persistent",
        "PersistentRel::insert");
        tuple->printon(Cor_exEnv->err_strm);
        return 0;
    }
#endif

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_indexing) {
            *(Cor_exEnv->trace_strm) << "PersistentRel::insert_raw: ";
            name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "(";
            tuple->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ") ";
        }
#    endif

    // Materialize a tuple in the desired format as an object to be inserted
    // insert it into the file
    count++;
    insertObject(tuple->args());

    delete &(tuple->args());
    delete tuple;
    return 1;
}

// *************************************************************************
int Cor_PersistentRel::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (there_are_marks()) {

        if (delta_indexed) {
            delta_relations->insert_tuple(tuple);
        } else {
            tlist.push(UnknownHashValue,tuple);
            count++;
        }
    } else {
        insert_raw(tuple);
    }

    return 1;
}

// *************************************************************************
Cor_StorageRelation *make_pers_relation(Cor_CArg& name, int arity,
    int delta_index,
    int set_subsum, Cor_ArgList *a_list = NULL)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized) {
        *(Cor_exEnv->err_strm) << "Not connected to EXODUS Server!\n";
        *(Cor_exEnv->err_strm) << "Unable to create disk relation!\n";
        return Cor_make_local_relation(name, arity, delta_index, set_subsum);
    }
    int error = 0;

    // Create a file descriptor for the new relation

    if (!a_list) {
        a_list = new Cor_ArgList(arity);
        Cor_CArg intsym = Cor_CArg("int");
        for (int i = 0; i < arity; i++)
            (*a_list)[i] = intsym;
    }

    Cor_CArg *f_arg = make_arg(name.symbol(),a_list);
    Cor_FileDescStruct *fd = new Cor_FileDescStruct(f_arg, error, 1 /*temp*/);
    f_arg->destroy();

    Cor_PersistentRel *p_rel = new Cor_PersistentRel(name, arity, error, fd);

    p_rel->delta_indexed = delta_index;
    p_rel->check_subsum = set_subsum;

    delete fd;

    if (error) {
        delete p_rel;
        return NULL;
    }
    
    return (Cor_StorageRelation *)p_rel;
}

// *************************************************************************
void deletePersistentRelation(Cor_CArg& name, int arity)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return;

    int input = 'Y';

    if (Cor_exEnv->C_interactive_mode_default) {
        *(Cor_exEnv->err_strm) << 
                "\nAre you sure you want to delete persistent relation ";
        *(Cor_exEnv->err_strm)
            << name.symbol() << " arity " << arity
            << " ? (Y/N)\n";
        input = getchar();
    }

    if ((input == 'Y') || (input == 'y')) {
        SM.deleteFile(name, arity);
    } else {
        *(Cor_exEnv->err_strm) << 
            "Delete of persistent relation aborted by user !\n";
    }
}

// *************************************************************************
void Cor_PersistentRel::empty_relation(int delete_tuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return;

    int e;
    OID oid, nextoid;
    OBJHDR objhdr;
    BOOL eof;

    if (!delete_tuples) return;

    e = sm_GetFirstOid(SM.bufGroup(), &(fd.fid), &oid, &objhdr, &eof);
    if (ErrorCheck(e, "sm_GetFirstOid") || eof) return;

    for (;;) {
        e = sm_GetNextOid(SM.bufGroup(), &oid, &nextoid, &objhdr, &eof);
        // If there are no more tuples, return.
        if (ErrorCheck(e, "sm_GetNextOid") || eof)
            return;

        sm_DestroyObject(SM.bufGroup(), &oid);
        oid = nextoid;
        // CHNG:wX
        // If there are no more tuples, return.
        if (ErrorCheck(e, "sm_GetNextOid") || eof)
            return;
    }
}

// *************************************************************************
void Cor_PersistentRel::release(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int e;

     // CHNG::wX
    if (!iterator.index) {
                                 /* if (((int)(iterator.index)) < 0) {*/
        // Full file scan 
        if (iterator.scanDesc) {
            e = sm_CloseScan((SCANDESC *)iterator.scanDesc);
            ErrorCheck(e, "sm_CloseScan");
            iterator.scanDesc = NULL;
        }

    } else {
        // index scan
        if (iterator.cursor) {
          //delete (SMCURSOR *)(iterator.cursor); CAUSES CORE DUMP on SPARC
          // CHNG:wX
          INVALIDATE_CURSOR(*((SMCURSOR *)iterator.cursor));
          iterator.cursor = NULL;
        }
    }
}

// *************************************************************************
int Cor_PersistentRel::start_bulk_load()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (!SM_initialized)
        return 0;

    if (lddesc) {
        *(Cor_exEnv->err_strm) << "previous bulk load not closed !\n";
        return 0;
    }

    int e;

    // NEW_CHNG:wX
    // int e = sm_OpenLoad(&(fd.fid), LOGICAL_LOAD, 8, 1.0, &lddesc);
    // *(Cor_exEnv->err_strm) << "lddesc is " << lddesc << endl;

    if (ErrorCheck(e, "sm_OpenLoad"))
        return 0;
    else
        return 1;
}

// *************************************************************************
int Cor_PersistentRel::end_bulk_load()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!SM_initialized)
        return 0;

    if (!lddesc) {
        *(Cor_exEnv->err_strm) << "bulk load not opened !\n";
        return 0;
    }

    *(Cor_exEnv->err_strm) << "lddesc is " << lddesc << endl;

    int e = sm_CloseLoad(lddesc);

    lddesc = NULL;

    if (ErrorCheck(e, "sm_CloseLoad")) return 0;
    else return 1;
}

// *************************************************************************

#endif WITH_PERSISTENCE
