// *************************************************************************
//    SortedRelation.C
// *************************************************************************

#include "SortedRelation.h"

#include "Util.h"
#include "Globals.h"
#include "CoralError.h"

//#include "Unify.h"
#include "TupleIterator.h"
#include "Stack.h"

#include "Debug.h"

//#define DEBUG_SORTED_REL 

// *************************************************************************
Cor_SortedRelation::Cor_SortedRelation(int arity, Cor_IntStack *sort_by,
        int incr_order) : Cor_StorageRelation(arity) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    _list = NULL;
    _sort_by = sort_by;
    _incr_order = incr_order;
}

// *************************************************************************
Cor_SortedRelation::~Cor_SortedRelation() 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    empty_relation(1);
    delete _sort_by;
}

// *************************************************************************
void Cor_SortedRelation::Init(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (it.ipos == Scanning_Init) {
        
        it.stack_mark.get_mark();
        it.ipos = Scanning_Known;
        it.SortNext = _list;
        it.SortPrev = NULL;

    } else {
        it.stack_mark.pop_to();
    }
}

// ************************************************************************
Cor_Tuple *Cor_SortedRelation::Done(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    it.ipos = Scanning_Done;
    it.set_no_match();

    return(NULL);
}

// *************************************************************************
// Standard 3-way comparison function between two nodes.
// Takes the order desired into account.
//             -1 : left comes before right (in a scan)
//              0 : they're equal (on the sort-by columns)
//              1 : right comes before left
// *************************************************************************
int Cor_SortedRelation::Compare(Cor_TupleNode *left, Cor_TupleNode *right)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, lim, cmp = -1;

    Cor_Tuple *ltup = left->tuple();
    Cor_Tuple *rtup = right->tuple();

    ASSERT(ltup);
    ASSERT(rtup);

    lim = _sort_by->count();

    for (j=0; j < lim; j++) {

        int argnum = _sort_by->get_ith(j);

        if (_incr_order) 
            cmp = ltup->args()[argnum].compare( &(rtup->args()[argnum]) );
        else
            cmp = rtup->args()[argnum].compare( &(ltup->args()[argnum]) );

          // Stop if they are not equal
        if (cmp)
            break;
    }

    return(cmp);
}

// ************************************************************************
Cor_Tuple *Cor_SortedRelation::get_next_tuple(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (Cor_exEnv->C_interrupt_raised)
      return(Done(it));

    Init(it);

    Cor_Tuple *tuple;

      // In case we need to delete the current one
    Cor_TupleNode *prev, *curr = NULL;

    do {
        if (it.SortNext) {
            prev = it.SortPrev;
            curr = it.SortNext;
            tuple = it.SortNext->tuple();
            it.SortPrev = it.SortNext;
            it.SortNext = it.SortNext->next();
        } else 
            return(Done(it));

        ASSERT(tuple);
        if (tuple->is_deleted()) {
            if (scan_count == 1) {
                if (prev) {
                    prev->set_next(it.SortNext);
                } else {
                    _list = it.SortNext;
                }
                delete curr->tuple();
                delete curr;
            }
        } else
            if (Cor_unify_literal(it, tuple, Cor_exEnv->dbg_get_next)) {
                it.reset_no_match();
                return(tuple);
            }

    } while(1);

    // Unreachable
    return(NULL);
}

// *************************************************************************
Cor_BindEnv *Cor_SortedRelation::get_next(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *result = get_next_tuple(it);

    if (!result)
        return(NULL);

    return(it.bindenv);
}

// *************************************************************************
int Cor_SortedRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef DEBUG_SORTED_REL
    cout << "Inserting tuple " << *tuple << " into relation " << this << endl;
#endif

    if (tuple->is_deleted())
        return(0);

    Cor_TupleNode *curr, *prev = NULL;
    Cor_TupleNode *new_node = new Cor_TupleNode(tuple);

    for (curr=_list; curr; curr = curr->next() ) {

          // Find spot to insert
        if (Compare(new_node, curr) <= 0) {
            new_node->set_next(curr);
            if (prev)
                prev->set_next(new_node);
            else
                _list = new_node;
            new_node = NULL; 
            break;
        }

        prev = curr;
    }

      // Insert at tail
    if (new_node) {
        if (prev)
            prev->set_next(new_node);
        else
            _list = new_node;
    }

    count++;

    return(1);
}

// *************************************************************************
void Cor_SortedRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, lim;

    *out << name.symbol() << "/" << arity() << "\t: (base)\t: (sorted by [";

    lim = _sort_by->count();
    for (j=0; j < lim; j++) {
        int argn = _sort_by->get_ith(j);
        *out << argn;
        if (j != lim-1)
            *out << ",";
    }
    
    *out << "] in " << (_incr_order ? "increasing" : "decreasing")
         << " order)\n";
}

// *************************************************************************
void Cor_SortedRelation::release(Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Done(it);
}

// *************************************************************************
/*
int is_subsumed(Cor_Tuple *tuple,
        Cor_RMark *startmark, Cor_RMark *endmark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
}
*/
// *************************************************************************
void Cor_SortedRelation::empty_relation(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_TupleNode *curr, *prev = NULL;

    for (curr=_list; curr; curr = curr->next() ) {
        if (deleteTuples)
            delete curr->tuple();
        delete prev;
        prev = curr;
    }

    delete prev;
    _list = NULL;
}

// *************************************************************************
