// *************************************************************************
//    ServerSolvers.C
// *************************************************************************

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "ServerSolvers.h"

#include "Config.h"
#include "Globals.h"
#include "CoralError.h"
#include "Interface.h"

#include "ArgList.h"
#include "BuiltinSymbols.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "BuiltinUtil.h"
#include "Util.h"

#include "Suspend.h"
#include "Suspension.h"
#include "ConnectArray.h"
#include "SockStream.h"
#include "Collector.h"
#include "ClientProtocol.h"

#include "Debug.h"

// *************************************************************************
// answer(n) -
//    This function changes the current user's num_answers.
//    This is the number of answers that will be returned for a query
//    before it is suspended.
// *************************************************************************
Cor_BindEnv *Cor_AnswerSolver(Cor_BuiltinRelation&,
                              Cor_TupleIterator& it)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (it.arg_list.count() != 1) {
        *(Cor_exEnv->out_strm)
            << "Answers per query presently at "
            << Cor_curUser->num_answers << endl;
    } else {

        int n = DEFLT_ANSWRS_PER_QRY;
        Cor_Term terms[1];

        if (Cor_CheckArgTypes(it.arg_list, 1, 1, "i", terms, 1, it.bindenv)
                == Coral_Args_Okay)
            n = terms[0].expr.NumIval();

        if (n <= 0)
            n = Cor_curUser->num_answers = DEFLT_ANSWRS_PER_QRY;
        else
            Cor_curUser->num_answers = n;

        *(Cor_exEnv->out_strm) << "Number of answers set to: " << n << endl;
    }

    return(it.bindenv);
}

// *************************************************************************
// login("user_name", "passwd") -
//    This function creates a new user on this connection with the login
//    name given.  A numerical user id (valid this session only) is
//    echoed back.  There is presently no validation of the password.
// *************************************************************************
Cor_BindEnv *Cor_LoginSolver(Cor_BuiltinRelation&,
                             Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 2) {
        *(Cor_exEnv->err_strm) << 
                  "login(user_name, password). Two arguments needed.\n";
        iterator.set_no_match();
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term(term2);
        if ((!term1.expr.is_symbol())
            || (!term2.expr.is_symbol())) {

            *(Cor_exEnv->err_strm) << 
                  "login() - user_name and password must be strings.\n";
            iterator.set_no_match();
            return (iterator.bindenv);
        }
        char *u_name = term1.expr.symbol();
        char *passwd = term2.expr.symbol();
        int id = Cor_curConn->NewUser(u_name, passwd);
        if (id < 0) {
            *(Cor_exEnv->err_strm) << "login failed.\n";
        } else {
            *(Cor_exEnv->out_strm) <<  CP_USER << id << "\n";
            *(Cor_exEnv->out_strm) <<  CP_LOGIN << u_name << "\n";
            *(Cor_exEnv->out_strm) <<  CP_PSWD << passwd << "\n";
            *(Cor_exEnv->out_strm) << "User " << u_name
                << " logged in. Session id: " << id << endl;
        }
    }
    return (iterator.bindenv);
}

// *************************************************************************
// logout(id) -
//    This function removes a user from the connection, freeing up space
//    for another user.  A friendly client should logout all of its
//    users before closing its connection, but if it doesn't coral
//    will take care of the clean up. 
// *************************************************************************
Cor_BindEnv *Cor_LogoutSolver(Cor_BuiltinRelation&,
                              Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int id;

    if (iterator.arg_list.count() == 0) {
        // Default - logout the current user
        id = Cor_curUser->user_id;

    } else if (iterator.arg_list.count() != 1) {

        *(Cor_exEnv->err_strm) << 
             "Usage: logout. logout(user_id). or logout(login_name).\n";
        return (iterator.bindenv);
    } else {

        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);

        if (term1.expr.is_symbol()) {

            char *nme = term1.expr.symbol();

            if ((id = Cor_curConn->FindUser(nme)) < 0) {
                *(Cor_exEnv->err_strm) << "User '" <<  nme << "' not found.\n";
                return (iterator.bindenv);
            }
        } else {
            id = make_int(term1.expr);
            if ((id < 0) || (id > MAX_USRS_PER_CONN)
                    || (!Cor_curConn->userArray[id])) {
                *(Cor_exEnv->err_strm) << "Invalid id: " <<  id << "\n";
                return (iterator.bindenv);
            }
        }
    }

    Cor_curConn->DeleteUser(id);
    return (iterator.bindenv);
}

// *************************************************************************
// who(). -
//    List all users logged in on all connections.
// *************************************************************************
Cor_BindEnv *Cor_WhoSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int cnt;

    if (iterator.arg_list.count() == 0) {

        cnt = Cor_clients->ListWho(Cor_exEnv->out_strm);
        if (!Cor_exEnv->C_quiet_mode_default)
            *(Cor_exEnv->out_strm) << "Number of users: " <<  cnt << "\n";

    } else {

        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        int conn = make_int(term1.expr);

        if ((conn < 0) || (conn > MAX_CONNECTIONS) || (!Cor_clients->connArray[conn]))
            *(Cor_exEnv->err_strm) << "Invalid connection: " <<  conn << "\n";

        else {
            cnt = Cor_clients->ListWho(Cor_exEnv->out_strm,
                    Cor_clients->connArray[conn]);
            if (!Cor_exEnv->C_quiet_mode_default)
                *(Cor_exEnv->out_strm)
                        << "Number of users on connection "
                        << conn << ": " << cnt << endl;
        }
    }
    return (iterator.bindenv);
}

// *************************************************************************
// whoami() -
//    Tell who the current user is.
// *************************************************************************
Cor_BindEnv *Cor_WhoamiSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 0) {
        *(Cor_exEnv->err_strm) << "whoami().: No arguments needed.\n";
    } else
        Cor_curUser->whoami(Cor_exEnv->out_strm);

    return (iterator.bindenv);
}

#ifdef DO_WODI
int to_wodi_flush = 0;
#endif

// *************************************************************************
// yield() -
//    Release the server to allow other connections a chance to use it.
// *************************************************************************
Cor_BindEnv *Cor_YieldSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 0)
        *(Cor_exEnv->err_strm) << "yield().: No arguments needed.\n";
    else {
        Cor_curUser->yield();
#ifdef DO_WODI
        to_wodi_flush = 1;
#endif
    }
    return(iterator.bindenv);
}

// *************************************************************************
// setqid() -
//    Set the query id to use for this user's queries
//    (until they set it to something else).
// *************************************************************************
Cor_BindEnv *Cor_SetQidSolver(Cor_BuiltinRelation&,
                Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    unsigned int qid = 0;

    if (iterator.arg_list.count() != 1) {
        *(Cor_exEnv->err_strm) << 
            "Usage: setqid(query_id).\n";
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);

        qid = make_int(term1.expr);
    }

    Cor_curUser->qid = qid;

    return(iterator.bindenv);
}

// *************************************************************************
// collection() -
//    Create/Join the collection for the relation specified.
// *************************************************************************
Cor_BindEnv *Cor_CollectionSolver(Cor_BuiltinRelation&,
                Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *relname = NULL;
    int   arity;

    if (iterator.arg_list.count() != 3) {
        *(Cor_exEnv->err_strm)
          << "Usage: collection(relname,arity,type).\n"
          << "    where type is 'producer' or 'consumer'.\n";
        iterator.set_no_match();
        return(NULL);
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_Term term2(iterator.arg_list[1], iterator.bindenv);
        Cor_Term term3(iterator.arg_list[2], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        Cor_deep_dereference_term(term2);
        Cor_deep_dereference_term(term3);

        if (   !term1.expr.is_symbol()
            || !term2.expr.is_int()
            || !term3.expr.is_symbol()  ) {
                *(Cor_exEnv->err_strm)
                  << "Usage: collection(relname,arity,type).\n"
                  << "    where type is 'producer' or 'consumer'.\n";
                iterator.set_no_match();
                return(NULL);
        } else {
            relname = term1.expr.symbol();
            arity = make_int(term2.expr);

            Cor_CU_Type cutype;

              // This will find this relation if it already exists.
            Cor_Relation *rel = Cor_make_relation(relname,arity);

            if (!rel->collector) {
              // Relation needs a Collector
                rel->collector = new Cor_Collector(rel);
            }

            if (term3.expr == *Cor_ConsumerSymbol)
                cutype = Cor_Consumer;
            else if (term3.expr == *Cor_ProducerSymbol)
                cutype = Cor_Producer;
            else {
                *(Cor_exEnv->err_strm)
                  << "Error: invalid type, assuming 'producer'.\n\n"
                  << "Usage: collection(relname,arity,type).\n"
                  << "    where type is 'producer' or 'consumer'.\n";
                cutype = Cor_Producer;
            }

            rel->collector->AddUser(Cor_curUser, cutype);
        }
    }

      // Need to yield to join the collection immediately
    Cor_curUser->yield();
#ifdef DO_WODI
    to_wodi_flush = 1;
#endif

    return(iterator.bindenv);
}

// *************************************************************************
// user(id) -
//    This function switches to the user specified, if they exist on this
//    connection.  If the user id specified is invalid, the user remains
//    unchanged.  A connection can always switch to id 0, which is user
//    "nobody".
// *************************************************************************
Cor_BindEnv *Cor_UserSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int id = 0;

    if (iterator.arg_list.count() != 1)
        *(Cor_exEnv->err_strm) << 
            "Usage: user(user_id). or user(login_name).\n";
    else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);

        if (term1.expr.is_symbol()) {

            char *nme = term1.expr.symbol();

            if ((id = Cor_curConn->FindUser(nme)) < 0) {
                *(Cor_exEnv->err_strm) << "User '" <<  nme << "' not found.\n";
                return (iterator.bindenv);
            }
        } else {

            id = make_int(term1.expr);
            if ((id < 0) || (id > MAX_USRS_PER_CONN)
                        || (!Cor_curConn->userArray[id])) {
                *(Cor_exEnv->err_strm) << "Invalid id: " <<  id << "\n";
                return (iterator.bindenv);
            }
        }

        Cor_curConn->useUser(id);
        // Update the client
        *(Cor_exEnv->out_strm) <<  CP_USER <<  id << "\n";
    }

    return(iterator.bindenv);
}

// *************************************************************************
// kill() -
//    Delete a suspended query.
// *************************************************************************
Cor_BindEnv *Cor_KillSolver(Cor_BuiltinRelation&,
                            Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 1) {
        *(Cor_exEnv->err_strm) << "syntax: kill(id). One argument needed.\n";
        iterator.set_no_match();
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        int id = make_int(term1.expr);
        Cor_suspension->kill(id);
    }

    return (iterator.bindenv);
}

// *************************************************************************
// list_sus() -
//    List suspended queries.
//    list_sus(user_id) for certain person(user_id>=0),
//    list_sus() for current user, list_sus(all) for all users on
//    this connection. 
// *************************************************************************
Cor_BindEnv *Cor_ListSusSolver(Cor_BuiltinRelation&,
                               Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int id;

    if (iterator.arg_list.count() == 0)
        Cor_suspension->list(Cor_curUser->user_id);        // default

    else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        if (term1.expr.is_symbol()) {
            char *s = term1.expr.symbol();

            if (!strcmp(s, "all"))
                Cor_suspension->list(-2);        // all users

            else if ((id = Cor_curConn->FindUser(s)) < 0)
                *(Cor_exEnv->err_strm) << "User '" << s << "' not found.\n";

            else
                Cor_suspension->list(id);

        } else {
            int id = make_int(term1.expr);
            if ((id < 0) || (id > MAX_USRS_PER_CONN)
                    || (!Cor_curConn->userArray[id]))
                *(Cor_exEnv->err_strm) << "Invalid id: " << id << endl;
            else
                Cor_suspension->list(id);
        }
    }
    return (iterator.bindenv);
}

// *************************************************************************
// resume() -
//    Resume a suspended query.
//    resume(sq_id) to resume, a user may only resume their own suspended
//    queries.
// *************************************************************************
Cor_BindEnv *Cor_ResumeSolver(Cor_BuiltinRelation&,
                              Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() != 1) {
        *(Cor_exEnv->err_strm) << 
                "Error: resume(id). One argnument needed.\n";
        iterator.set_no_match();
        return (iterator.bindenv);
    } else {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);
        int id = make_int(term1.expr);
        // *(Cor_exEnv->out_strm) << "going to resume " << id << endl;
        Cor_suspension->resume(id);
        return (iterator.bindenv);
    }
}

// *************************************************************************
// sync() -
//    Print out a sync message for the client.  Sync with a single argument
//    prints out that argument too.
// *************************************************************************
Cor_BindEnv *Cor_SyncSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator & iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (iterator.arg_list.count() == 1) {
        Cor_Term term1(iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term(term1);

        if (term1.expr.is_symbol()) {

            char *s = term1.expr.symbol();
            *(Cor_exEnv->out_strm) << endl <<  CP_SYNC << s << endl;

        } else {

            int id = make_int(term1.expr);
            *(Cor_exEnv->out_strm) << endl <<  CP_SYNC << id << endl;
        }
    } else {
        *(Cor_exEnv->out_strm) << endl <<  CP_SYNC << endl;
    }

    return (iterator.bindenv);
}

// *************************************************************************
