// *************************************************************************
//    Adorn.C
// *************************************************************************

#include "Adorn.h"

#include "Print.h"
#include "SipGen.h"

#include "Globals.h"
#include "Annotations.h"
#include "RelationMisc.h"        // has_grouping_args
#include "ExportInfo.h"
#include "Sips.h"
#include "Unparse.h"
#include "Interface.h"
#include "CArg.h"

// *************************************************************************

static int group_rule_num = 0;
static int adorn_type = 0;

// *************************************************************************

static void deepen_adorn(Cor_Collection *pre_ad_rules,
    Cor_Collection *ad_rules, Cor_CArg& name, Cor_BitVector *adorn);
static int adorn_rule(Cor_Collection *pre_ad_rules, Cor_Clause *rule,
    Cor_Collection *ad_rules, Cor_CArg& name, Cor_BitVector *adorn);
static void adorn_rule_preds(Cor_Collection *pre_ad_rules,
    Cor_Clause *adrule, int p, Cor_Literal *pred,
    Cor_Collection *ad_rules);

static int isbound(Cor_CArg *a, Cor_BitVector *u);
static int equals_pred(Cor_Literal *lit1, Cor_Literal *lit2);


// SEE NOTE!
static Cor_Literal *copy_pred(Cor_Literal *prototype);

// *************************************************************************
/*
    Construct the adorned system.  adorn_type==1: construct as described in
    "On the Power of Magic", adorn_type==2: as in "Magic Templates:
    A Spellbinding Approach to Logic Programs".

    Type 2 is different from type 1 as follows:
        - bound query args are any arg position that contains a function
            symbol, constant, or repeated variable.
        - to compute the adornment of a derived predicate, an argument is
            considered bound if *some *variable appearing in it is in the
            union of sip labels.
*/

int adorn(Cor_Collection *pre_ad_rules, Cor_Collection *ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BitVector adorn;

    if (Cor_parserStruct->CurModule->ExportedPreds == NULL) {
        *cerr_strm << "'export' missing or bad format.\n";
        return 0;
    }
    else {
        for (Cor_ExportInfo *export_info =
                Cor_parserStruct->CurModule->ExportedPreds;
                export_info != NULL; export_info = export_info->next)
        {
            deepen_adorn(pre_ad_rules, ad_rules,
                    export_info->pred_name, export_info->adornment);
        }
    }

    return 1;
}

// *************************************************************************
static void deepen_adorn(Cor_Collection *pre_ad_rules,
    Cor_Collection *ad_rules, Cor_CArg& name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *rule;

    ad_rules->use_hash_table(1);

    for (rule = ad_rules->get_match(name); rule; rule = rule->link) {
        if (rule->head->adorn == *adorn)
            return;
    }

    // Find rules that match name and adorn.
    pre_ad_rules->use_hash_table(1);

    for (rule = pre_ad_rules->get_match(name); rule; rule = rule->link) {
        if (!adorn_rule(pre_ad_rules, rule, ad_rules, name, adorn))
            break;
    }

    // Update the annotations
    Cor_update_annotations(name, *adorn);
}

// *************************************************************************
static int adorn_rule(Cor_Collection *pre_ad_rules, Cor_Clause *rule,
    Cor_Collection *ad_rules, Cor_CArg& name, Cor_BitVector *adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Sips *s;

    // Build the skeleton of the new adorned rule
    Cor_Clause *adrule = Cor_copy_rule(rule);
    adrule->head = copy_pred(rule->head);
    adrule->head->base_pred = 0;
    if (!adorn)
        adorn = new Cor_BitVector();
    adrule->head->adorn = *adorn;
    Cor_alloc_preds(adrule, rule->num_preds);

    // attach only the appropriate sip
    adrule->sips = new Cor_Sips();
    adrule->sips->adorn = *adorn;
    ad_rules->push(adrule);
    adrule->sips->next = NULL;

    // For this rule, find its sip with the same adornment as the
    // worklist predicate
    int sips_complete = 1;
    Cor_Sip *asip;
    for (s=rule->sips; s; s=s->next) {
        if (s->adorn == *adorn) {
            for (int pred_num=0; pred_num < rule->num_preds; pred_num++) {
                for (asip=s->sip_list;asip;asip=asip->next) {
                    if (asip->head==pred_num+1)
                        break;
                }
                if (!asip)
                    sips_complete = 0;
            }
            break;
        }
    }

    if (s != NULL && sips_complete)
        adrule->sips->sip_list = s->sip_list;
    else {
        if (rule->head->args.count() > COR_MAX_ARGLIST_SIZE) {
            *cerr_strm << "Error: Rule (line " << rule->line_number
                << ") has too many args (" << rule->head->args.count()
                << ") - ignored\n" << flush;
            return(1);
        }

        ClauseInfo rule_info[1];
        setup_vars(rule, rule_info);
        gen_sip_rule(rule_info, *adorn);
        cleanup_vars(rule_info);
        for (s=rule->sips; s; s=s->next) {
            if (s->adorn == *adorn)
                 break;
        }

        if (!s) {
            *cerr_strm << "Internal Error: no sip matching adornment ";
            Cor_print_adornment(adorn, cerr_strm);
            *cerr_strm << " for " << name.symbol()
                << " (line " << rule->line_number << ")\n";
            return 0;
        }
        adrule->sips->sip_list = s->sip_list;
    }

    // Now compute the adornment for each derived predicate in the
    // rule's right-hand-side.
    for (int p=0; p<rule->num_preds; ++p) {
        adorn_rule_preds(pre_ad_rules, adrule, p, rule->preds[p], ad_rules) ;
    }

    // Now handle any head deletes
    adrule->head_deletes_count = rule->head_deletes_count;
    adrule->head_deletes = new Cor_Literal* [adrule->head_deletes_count];
    for (int hdnum=0; hdnum < rule->head_deletes_count; hdnum++) {
        adrule->head_deletes[hdnum] = copy_pred(rule->head_deletes[hdnum]);
        adrule->head_deletes[hdnum]->base_pred = 0;
        // So adorn will get printed

        int rulepos;
        for (rulepos=0; rulepos < rule->num_preds; rulepos++) {
            if (equals_pred(rule->preds[rulepos], rule->head_deletes[hdnum]))
                break;
        }
        if (rulepos < rule->num_preds) {
            adrule->head_deletes[hdnum]->adorn = adrule->preds[rulepos]->adorn;
        }
    }
    return 1;
}

// *************************************************************************
static void adorn_rule_preds(Cor_Collection *pre_ad_rules,
        Cor_Clause *adrule, int p, Cor_Literal *pred,
        Cor_Collection *ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Sip *arc;

    //    Add the predicate to the adorned rule
    adrule->preds[p] = copy_pred(pred);
    if (pred->base_pred) {
        adrule->preds[p]->adorn.setlen(0);
    }
    else {
        // Compute the union of labels of the sip arcs entering the node
        // labelled by pred
        Cor_BitVector u;
        for (arc=adrule->sips->sip_list; arc; arc=arc->next) {
            // sub 1 since p indexes rhs only (not head)
            if (arc->head-1 == p) {
                u.or_bits(&(arc->bound_vars));
            }
        }

        // For each arg in pred, if all vars in that arg are in the
        // union, mark the arg as bound
        int i = 0;
        FOR_EACH_ARG(a, pred->args) {
            if (isbound(a, &u)) {
                adrule->preds[p]->adorn.set(i);
            }
            i++;
        } END_EACH_ARG

        // Find best allowed adornment
        set_best_allowed_adorn(adrule->preds[p]);
        deepen_adorn(pre_ad_rules, ad_rules, pred->pred,
                    &(adrule->preds[p]->adorn));
    }
}

// *************************************************************************
//  Type 1:
//      Recursively traverse the given argument and return 1 if all
//      sub-args are in the set of variables represented by the BitVector u
//
//  Type 2:
//      Recursively traverse the given argument and return 1 if there is at
//      least one sub-arg that is in the set of variables represented by the
//      BitVector u
// *************************************************************************

static int isbound(Cor_CArg *a, Cor_BitVector *u)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (adorn_type == 2) {
        switch (a->Kind()) {

        case VAR_ARG:
            return(u->test(a->var_num()));

        case SYMBOL_ARG:
        case INT_ARG:
        case DOUBLE_ARG:
            return(1);

        case DONT_CARE_ARG:
            return(0);

        case LIST_ARG:
            if (isbound(a->car(),u))
                return(1);
            if (isbound(a->cdr(),u))
                return(1);
            return(0);

        case FUNCT_ARG:
            FOR_EACH_ARG(arg, a->args()) {
                if (isbound(arg, u))
                    return(1);
            } END_EACH_ARG
            return(0);

        case MAX_GRPING_ARG:
        case MIN_GRPING_ARG:
        case ANY_GRPING_ARG:
            return(0);

        default:
            break;
        }
    } else {

        switch (a->Kind()) {
        case VAR_ARG:
            return(u->test(a->var_num()));

        case SYMBOL_ARG:
        case INT_ARG:
        case DOUBLE_ARG:
            return(1);

        case DONT_CARE_ARG:
            return(0);

        case LIST_ARG:
            if (! isbound(a->car(),u))
                return(0);
            if (! isbound(a->cdr(),u))
                return(0);
            return(1);

        case FUNCT_ARG:
            FOR_EACH_ARG(arg, a->args()) {
                if (!isbound(arg, u))
                    return(0);
            } END_EACH_ARG
            return(1);

        case MAX_GRPING_ARG:
        case MIN_GRPING_ARG:
        case ANY_GRPING_ARG:
            return(0);

        default:
            break;
        }
    }
    return(1);
}

// *************************************************************************
static int equals_pred(Cor_Literal *lit1, Cor_Literal *lit2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!(lit1->name().equals(lit2->name()))
                || lit1->arity() != lit2->arity())
        return(0);

    for (int i=0; i < lit1->arity(); i++) {
        if (!lit1->args[i].equals(lit2->args[i]))
            return 0;
    }
    return(1);
}

// *************************************************************************
void set_best_allowed_adorn(Cor_Literal *lit)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur = Cor_parserStruct->AllowedAdornList;

    int best_shortfall = 1000;   // Infinity
    Cor_BitVector *best_adorn = NULL;
    int restrictions = 0;        // Adornments restricted for this predicate

    for (;cur != NULL; cur = cur->next()) {
        if (cur->pred == lit->predicate->name) {

            int i = lit->adorn.len();
            if (cur->adorn->len() != i)
                continue;
            // An adornment restriction has been specified for this predicate
            restrictions = 1;
            int shortfall = 0;
            while (--i >= 0) {
                if (!lit->adorn.test(i) && cur->adorn->test(i))
                    break; // Allowed adorn overbound
                if (lit->adorn.test(i) && !cur->adorn->test(i))
                    shortfall++;
            }
            if (i >= 0) {
                // Allowed adorn is overbound  - ignore it
                continue;
            }
            if ((shortfall < best_shortfall) &&
                (!is_disallowed_adorn(cur->pred, *(cur->adorn))))
            {
                best_adorn = cur->adorn;
            }
        }
    }

    if (!restrictions)
        return;

    if (!best_adorn) {
        *cerr_strm
            << "Warning: adornment for " << lit->predicate->name.symbol()
            << " does not match any allowed adornment.\n"
            << "                -- adornment restrictions ignored.\n";
        return;
    }

    lit->adorn.init(*best_adorn);
    return;
}

// *************************************************************************
int is_disallowed_adorn(Cor_CArg& name, Cor_BitVector &adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_PredAnnotations *cur = Cor_parserStruct->DisallowedAdornList;

    int adorn_length = adorn.len();

    for(; cur != NULL; cur = cur->next()) {
        if (cur->pred == name) {
            if (cur->adorn->len() != adorn_length)
                continue;
            else if (*(cur->adorn) == adorn)
                return 1;
        }
    }
    return 0;
}

// *************************************************************************
/*
    The following code is to ensure that a rule with grouping in the head
    and more than one body predicate is split into two rules as follows:
        p(tbar, <Y>) :- l1, l2, ... ln.  is split up as
        (1): p(Xbar, <Y>) :- pprime(Xbar, Y).  and
        (2): pprime(tbar, Y) :- l1, l2, ... ln.
    This is to ensure correctness of the evaluation of the magic rewritten
    program.  It also ensures that grouping args are not bound in the
    adorned program.  Consequently, magic works for programs with grouping.
*/

void preprocess_for_grouping(Cor_Collection *orig_rules,
    Cor_Collection *pre_ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k, nrules;
    char *disallowed_head_adorn, *possible_allowed_head_adorn;
    Cor_Clause *preadrule, *rule;
    Cor_CArg new_pred_name;
    int new_pred_arity;


    group_rule_num = 0;

    nrules = orig_rules->count();
    for (k=0; k < nrules; k++) {
        rule = orig_rules->get_ith(k);

        Cor_Literal *head = rule->head;
        if (head->args.has_grouping_args()) {
            if (rule->num_preds == 0) {
                *cerr_strm << "ERROR: Cannot have grouping rule with "
                    << "an empty body.\n";
                return;
            }
            else {
                // valid grouping rule in the program
                // FIRST emit grouping argument restrictions
                disallowed_head_adorn = init_any_adorn(head->args.count());
                possible_allowed_head_adorn=init_any_adorn(head->args.count());

                int i = 0;
                FOR_EACH_ARG(a, head->args) {
                    if (a->is_grouping()) {
                        disallowed_head_adorn[i] = 'b';
                        possible_allowed_head_adorn[i] = 'f';
                    }
                    i++;
                } END_EACH_ARG

                Cor_add_disallowed_adornment(Cor_parserStruct, head->pred,
                    disallowed_head_adorn, head->args.count());
                Cor_add_allowed_adornment(Cor_parserStruct, head->pred,
                    possible_allowed_head_adorn, head->args.count());
                delete disallowed_head_adorn;
                delete possible_allowed_head_adorn;
                
                // NEXT rewrite grouping rules into two rules
                if (rule->num_preds > 1) {
                    new_pred_name = build_ad_grouped_pred_name(head);
                    new_pred_arity = grouped_pred_arity(head);
                    create_grouping_rule_1(rule, new_pred_name, new_pred_arity,
                        pre_ad_rules);
                    create_grouping_rule_2(rule, new_pred_name, new_pred_arity,
                        pre_ad_rules);
                    continue;
                }
            }
        }

        // Either there are no grouping args in the head or the body
        // had exactly one predicate -- the same thing needs to be done.
        // Just copy out the rule.
        preadrule = Cor_copy_rule(rule);
        preadrule->head = rule->head;
        Cor_alloc_preds(preadrule, rule->num_preds);
        for (j=0; j <rule->num_preds; j++) {
            preadrule->preds[j] = rule->preds[j];
        }
        preadrule->head_deletes_count = rule->head_deletes_count;
        preadrule->head_deletes = rule->head_deletes;

        pre_ad_rules->push(preadrule);
    }
}

// *************************************************************************
Cor_CArg& build_ad_grouped_pred_name(Cor_Literal *head)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *head_name = head->pred.symbol();
    int head_name_length = head->pred.symlen();

    char *buf = new char [head_name_length+10];
    group_rule_num++; // this gives it the unique numbering

    sprintf(buf, "grp_%d_",group_rule_num);
    char *buf1 = buf + strlen(buf);
    strcpy(buf1, head_name);
    buf1[head_name_length] = '\0';

      // LEAK
    Cor_CArg *tmp = new Cor_CArg(buf);
    return(*tmp);
}

// *************************************************************************
int grouped_pred_arity(Cor_Literal *head)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int ret_val = 0;
    FOR_EACH_ARG(a, head->args) {
        if (! a->is_grouping()) {
            ret_val++;
        } else {
            // it is a grouping argument
            ret_val += a->nested_args()->arity();
        }
    } END_EACH_ARG

    return(ret_val);
}

// *************************************************************************
char *init_any_adorn(int num)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char *ret_val = new char[num+1];

    for (int i = 0; i < num; i++)
        ret_val[i] = 'a';
    ret_val[num] = '\0';
    return ret_val;
}

// *************************************************************************
/*
    This adds the rule of the form
        p(Xbar, <Y>) :- pprime(Xbar, Y)
    obtained from
        p(tbar, <Y>) :- l1, l2, ... ln
*/

void create_grouping_rule_1(Cor_Clause *rule, Cor_CArg& new_pred_name,
    int new_pred_arity, Cor_Collection *pre_ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Clause *grouping_rule_1 = Cor_copy_rule(rule);
    Cor_Literal *head = rule->head;
    int j, k;
    Cor_CArg *arg_ptr;
    Cor_CArg *new_groupingarg;

    // create the new head literal
    Cor_Literal *new_head = new Cor_Literal(Cor_parserStruct, head->pred,
        head->args.count());
    j = 0;
    arg_ptr = & (new_head->args[0]);
    FOR_EACH_ARG(a, head->args) {
        if (! a->is_grouping()) {
            Cor_CArg *carg = make_var(j++);
            *arg_ptr++ = *carg;
            carg->destroy();
        }
        else {
            // it was a grouping argument earlier and it has to be now
            new_groupingarg = new Cor_CArg();
            new_groupingarg->set_grping(a->grp_arity());
            new_groupingarg->set_grp_kind(a->Kind());

            for (k=0; k < new_groupingarg->grp_arity(); k++) {
                Cor_CArg *carg = make_var(j++);
                (new_groupingarg->nested_args()->args())[k] = *carg;
                carg->destroy();
            }

            *arg_ptr++ = *new_groupingarg;
            delete new_groupingarg;
        }
    } END_EACH_ARG

    // allocate the head
    grouping_rule_1->head = new_head;

    // create the new body literal
    Cor_Literal *new_body = new Cor_Literal(Cor_parserStruct, new_pred_name,
        new_pred_arity);
    arg_ptr = & (new_body->args[0]);
    for (j=0; j < new_pred_arity; j++) {
        Cor_CArg *carg = make_var(j);
        *arg_ptr++ = *carg;
        carg->destroy();
    }

    // allocate the body
    Cor_alloc_preds(grouping_rule_1, 1);
    grouping_rule_1->preds[0] = new_body;

    pre_ad_rules->push(grouping_rule_1);
}

// *************************************************************************
/*
    This implicitly adds the rule of the form
        pprime(tbar, Y) :- l1, l2, ... ln.
    obtained from
        p(tbar, <Y>) :- l1, l2, ... ln
*/

void create_grouping_rule_2(Cor_Clause *rule, Cor_CArg& new_pred_name,
    int new_pred_arity, Cor_Collection *pre_ad_rules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    Cor_Clause *grouping_rule_2;
    Cor_CArg *arg_ptr;
    Cor_Literal *head = rule->head;

    grouping_rule_2 = Cor_copy_rule(rule);

    // create the new head
    Cor_Literal *new_head = new Cor_Literal(Cor_parserStruct,
                                     new_pred_name, new_pred_arity);
    arg_ptr = &(new_head->args[0]);
    FOR_EACH_ARG(a, head->args) {
        if (! a->is_grouping()) {
            *arg_ptr++ = *a;
        } else {
              // it is a grouping argument
            for (j=0; j < a->grp_arity(); j++) {
                *arg_ptr++ = (a->nested_args()->args())[j];
            }
        }
    } END_EACH_ARG

    // allocate the head
    grouping_rule_2->head = new_head;

    // copy out the body
    Cor_alloc_preds(grouping_rule_2, rule->num_preds);
    for (j=0; j <rule->num_preds; j++) {
        grouping_rule_2->preds[j] = rule->preds[j];
    }

    pre_ad_rules->push(grouping_rule_2);
}

// *************************************************************************
// NOTE: THIS NEEDS TO BE FIXED
// ADD A MEANS OF COPYING A LITERAL TO LITERAL.Ch in CORAL

static Cor_Literal *copy_pred(Cor_Literal *prototype)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Literal *pred = new Cor_Literal(prototype->args.count());
    *pred = *prototype;
    pred->adorn.init(prototype->adorn); // Fake copy constructor.
    return pred;
}

// *************************************************************************

